<?php

namespace App\Doctrine;

use App\Entity\User;
use Doctrine\Common\EventSubscriber;
use Doctrine\ORM\Event\LifecycleEventArgs;
use Symfony\Component\Security\Core\Encoder\UserPasswordEncoderInterface;

/**
 * Automatically encrypt user password
 *
 * This is a Doctrine event listener, to ensure the password is *always* encrypted
 *
 * @package App\Doctrine
 * @see     User
 *
 */
class CryptPasswordListener implements EventSubscriber {
    /**
     * @var UserPasswordEncoderInterface The user's password encoder service
     */
    private UserPasswordEncoderInterface $passwordEncoder;

    /**
     * Class constructor
     *
     * Called using auto-wiring
     *
     * @param UserPasswordEncoderInterface $passwordEncoder The user's password encoder service
     */
    public function __construct (UserPasswordEncoderInterface $passwordEncoder) {
        $this->passwordEncoder = $passwordEncoder;
    }

    /**
     * @inheritDoc
     */
    public function getSubscribedEvents (): array {
        return [
            'prePersist',
            'preUpdate',
        ];
    }

    /**
     * When entity is inserted in database
     *
     * Automatically called by Doctrine
     *
     * @param LifecycleEventArgs $eventArgs The event arguments
     */
    public function prePersist (LifecycleEventArgs $eventArgs): void {
        // Check if entity is a User, else do nothing
        $entity = $eventArgs->getEntity();
        if (!$entity instanceof User) {
            return;
        }

        $this->encodePassword($entity);
    }
    /**
     * When entity is updated in database
     *
     * Automatically called by Doctrine
     *
     * @param LifecycleEventArgs $eventArgs The event arguments
     */
    public function preUpdate (LifecycleEventArgs $eventArgs): void {
        // Check if entity is a User, else do nothing
        $entity = $eventArgs->getEntity();
        if (!$entity instanceof User) {
            return;
        }

        $this->encodePassword($entity);

        // Force update of entity
        $doctrine = $eventArgs->getEntityManager();
        $doctrine->getUnitOfWork()->recomputeSingleEntityChangeSet(
            $doctrine->getClassMetadata(get_class($entity)),
            $entity
        );
    }

    /**
     * Perform teh password encryption (if necessary)
     *
     * @param User $user The user
     */
    private function encodePassword (User $user): void {
        // Mot de passe modifié ?
        if (empty($user->getPlainPassword())) {
            return;
        }

        // Crypt the password
        $user->setPassword(
            $this->passwordEncoder->encodePassword(
                $user,
                $user->getPlainPassword()
            )
        );
    }
}