<?php
/** @noinspection PhpPrivateFieldCanBeLocalVariableInspection */

namespace App\Entity;

use DateTimeImmutable;
use DateTimeInterface;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\Common\Collections\Collection;
use Doctrine\ORM\Mapping as ORM;
use InvalidArgumentException;
use Symfony\Bridge\Doctrine\Validator\Constraints\UniqueEntity;
use Symfony\Component\HttpFoundation\File\File;
use Symfony\Component\Security\Core\User\UserInterface;
use Symfony\Component\Validator\Constraints as Assert;
use Vich\UploaderBundle\Entity\File as VichEmbeddedFile;
use Vich\UploaderBundle\Mapping\Annotation as Vich;

/**
 * User entity
 *
 * @ORM\Entity(repositoryClass="App\Repository\UserRepository")
 * @UniqueEntity("email")
 * @Vich\Uploadable()
 */
class User implements UserInterface {
    /**
     * @var int The internal ID of user
     *
     * @ORM\Id()
     * @ORM\GeneratedValue()
     * @ORM\Column(type="integer")
     */
    private int $id;

    /**
     * @var string The user email
     *
     * @ORM\Column(type="string", length=180, unique=true)
     * @Assert\Email(mode="html5", message="user.invalid.email")
     */
    private string $email;

    /**
     * @var array The list of user's roles
     *
     * @ORM\Column(type="json")
     */
    private array $roles = [];

    /**
     * @var string The hashed password
     *
     * @ORM\Column(type="string")
     */
    private string $password;
    /**
     * @var string The "clear" password.
     *             Only used when changing the password
     *
     * @Assert\NotBlank(message="user.invalid.password")
     */
    private string $plainPassword;

    /**
     * @var string The user name
     *
     * @ORM\Column(type="string", length=255)
     * @Assert\NotBlank(message="user.invalid.name")
     */
    private string $name;
    /**
     * @var string|null The user first name
     *
     * @ORM\Column(type="string", length=255, nullable=true)
     */
    private ?string $firstname;
    /**
     * @var string|null The user gender
     *  - male
     *  - femalle
     *  - genderqueer
     *
     * @ORM\Column(type="string", length=25, nullable=true)
     * @Assert\Choice(callback="getGenders", message="user.invalid.gender")
     */
    private ?string $gender;

    /**
     * @var string The user language code (ISO 639-1)
     *
     * @ORM\Column(type="string", length=2)
     * @Assert\NotBlank(message="user.invalid.language")
     * @Assert\Language(message="user.invalid.language")
     */
    private string $language;

    /**
     * @var VichEmbeddedFile|null The user avatar
     *
     * @ORM\Embedded(class="Vich\UploaderBundle\Entity\File")
     */
    private ?VichEmbeddedFile $avatar;
    /**
     * @var File|null The user avatar "file" (NOT the mapped field)
     *
     * @Vich\UploadableField(
     *     mapping="user_avatar",
     *     fileNameProperty="avatar.name",
     *     size="avatar.size",
     *     mimeType="avatar.mimeType",
     *     originalName="avatar.originalName",
     *     dimensions="avatar.dimensions"
     * )
     */
    private ?File $avatarFile;
    /**
     * @var DateTimeInterface|null Date-Time of last update
     * Ensure Doctrine event listener is triggered if only avatar change
     *
     * @ORM\Column(type="datetime")
     */
    private ?DateTimeInterface $updatedAt;

    /**
     * @var Collection The user {@see User friends}
     *
     * @ORM\ManyToMany(targetEntity="App\Entity\User")
     * @ORM\JoinTable(
     *     name="friend",
     *     joinColumns={
     *          @ORM\JoinColumn(name="user_id", referencedColumnName="id")
     *     },
     *     inverseJoinColumns={
     *          @ORM\JoinColumn(name="friend_id", referencedColumnName="id")
     *     }
     * )
     */
    private Collection $friends;
    /**
     * @var Collection The user {@see ConfigNotification configured notifications}
     *
     * @ORM\OneToMany(targetEntity="App\Entity\ConfigNotification", mappedBy="user", orphanRemoval=true)
     */
    private Collection $configNotifications;
    /**
     * @var Collection Thue user {@see UserNotification notifications}
     *
     * @ORM\OneToMany(targetEntity="App\Entity\UserNotification", mappedBy="user", orphanRemoval=true)
     */
    private Collection $notifications;
    /**
     * @var Collection The user owned {@see Wish wishes}
     *
     * @ORM\OneToMany(targetEntity="App\Entity\Wish", mappedBy="owner")
     */
    private Collection $ownedWishes;
    /**
     * @var Collection The user {@see Participant participations}
     *
     * @ORM\OneToMany(targetEntity="App\Entity\Participant", mappedBy="user", orphanRemoval=true)
     */
    private Collection $participations;
    /**
     * @var Collection The user {@see Comment comments}
     *
     * @ORM\OneToMany(targetEntity="App\Entity\Comment", mappedBy="user", orphanRemoval=true)
     */
    private Collection $comments;

    public function __construct () {
        $this->email = '';
        $this->password = '';
        $this->plainPassword = '';

        $this->name = '';
        $this->firstname = null;
        $this->gender = null;

        $this->language = '';
        $this->avatar = new VichEmbeddedFile();
        $this->avatarFile = null;
        $this->updatedAt = new DateTimeImmutable();

        $this->friends = new ArrayCollection();
        $this->configNotifications = new ArrayCollection();
        $this->notifications = new ArrayCollection();
        $this->ownedWishes = new ArrayCollection();
        $this->participations = new ArrayCollection();
        $this->comments = new ArrayCollection();
    }
    public function getId (): ?int {
        return $this->id;
    }

    public function getEmail (): ?string {
        return $this->email;
    }
    public function setEmail (string $email): self {
        $this->email = $email;

        return $this;
    }

    /**
     * A visual identifier that represents this user.
     *
     * @see UserInterface
     */
    public function getUsername (): string {
        return (string)$this->email;
    }

    /**
     * @see UserInterface
     */
    public function getRoles (): array {
        $roles = $this->roles;

        // Every user must have at least the ROLE_USER
        $roles[] = 'ROLE_USER';

        return array_unique($roles);
    }
    public function setRoles (array $roles): self {
        $this->roles = $roles;

        return $this;
    }

    /**
     * @see UserInterface
     */
    public function getPassword (): string {
        return (string)$this->password;
    }
    public function setPassword (string $password): self {
        $this->password = $password;

        return $this;
    }

    public function getPlainPassword (): string {
        return (string)$this->plainPassword;
    }
    public function setPlainPassword (string $password): self {
        $this->plainPassword = $password;
        $this->password = '';                   // Force update of user

        // The crypted password will be handled by Doctrine\CryptPasswordListener

        return $this;
    }

    /**
     * @see UserInterface
     */
    public function getSalt () {
        // Security.yaml : use "bcrypt" which provide his own salt generator => nothing to do here
    }
    /**
     * @see UserInterface
     */
    public function eraseCredentials () {
        // Reset sensitive data, like clear password !
        $this->plainPassword = '';
    }

    public function getName (): ?string {
        return $this->name;
    }
    public function setName (string $name): self {
        $this->name = $name;

        return $this;
    }
    public function getFirstname (): ?string {
        return $this->firstname;
    }
    public function setFirstname (?string $firstname): self {
        $this->firstname = $firstname;

        return $this;
    }

    public static function getGenders (): array {
        return [
            'male',
            'female',
            'genderqueer',
        ];
    }
    public function getGender (): ?string {
        return $this->gender;
    }
    public function setGender (?string $gender): self {
        $this->gender = $gender;

        return $this;
    }

    public function getAvatar (): ?VichEmbeddedFile {
        return $this->avatar;
    }
    public function setAvatar (VichEmbeddedFile $avatar): self {
        $this->avatar = $avatar;

        return $this;
    }
    public function getAvatarFile (): ?File {
        return $this->avatarFile;
    }
    public function setAvatarFile (?File $avatar): self {
        $this->avatarFile = $avatar;
        if ($avatar !== null) {
            $this->updatedAt = new DateTimeImmutable();
        }

        return $this;
    }

    public function getLanguage (): string {
        return $this->language;
    }
    public function setLanguage (string $language): self {
        if (mb_strlen($language, 'UTF-8') != 2) {
            throw new InvalidArgumentException('The language code must have exactly 2 characters');
        }

        $this->language = $language;

        return $this;
    }

    /**
     * @return Collection|self[]
     */
    public function getFriends (): Collection {
        return $this->friends;
    }
    public function addFriend (self $friend): self {
        if (!$this->friends->contains($friend)) {
            $this->friends[] = $friend;
        }

        return $this;
    }
    public function removeFriend (self $friend): self {
        if ($this->friends->contains($friend)) {
            $this->friends->removeElement($friend);
        }

        return $this;
    }

    /**
     * @return Collection|ConfigNotification[]
     */
    public function getConfigNotifications (): Collection {
        return $this->configNotifications;
    }
    public function addConfigNotification (ConfigNotification $configNotification): self {
        if (!$this->configNotifications->contains($configNotification)) {
            $this->configNotifications[] = $configNotification;
            $configNotification->setUser($this);
        }

        return $this;
    }
    public function removeConfigNotification (ConfigNotification $configNotification): self {
        if ($this->configNotifications->contains($configNotification)) {
            $this->configNotifications->removeElement($configNotification);
            // set the owning side to null (unless already changed)
            if ($configNotification->getUser() === $this) {
                $configNotification->setUser(null);
            }
        }

        return $this;
    }

    /**
     * @return Collection|UserNotification[]
     */
    public function getNotifications (): Collection {
        return $this->notifications;
    }
    public function addNotification (UserNotification $notification): self {
        if (!$this->notifications->contains($notification)) {
            $this->notifications[] = $notification;
            $notification->setUser($this);
        }

        return $this;
    }
    public function removeNotification (UserNotification $notification): self {
        if ($this->notifications->contains($notification)) {
            $this->notifications->removeElement($notification);
            // set the owning side to null (unless already changed)
            if ($notification->getUser() === $this) {
                $notification->setUser(null);
            }
        }

        return $this;
    }

    /**
     * @return Collection|Wish[]
     */
    public function getOwnedWishes (): Collection {
        return $this->ownedWishes;
    }
    public function addOwnedWish (Wish $ownedWish): self {
        if (!$this->ownedWishes->contains($ownedWish)) {
            $this->ownedWishes[] = $ownedWish;
            $ownedWish->setOwner($this);
        }

        return $this;
    }
    public function removeOwnedWish (Wish $ownedWish): self {
        if ($this->ownedWishes->contains($ownedWish)) {
            $this->ownedWishes->removeElement($ownedWish);
            // set the owning side to null (unless already changed)
            if ($ownedWish->getOwner() === $this) {
                $ownedWish->setOwner(null);
            }
        }

        return $this;
    }

    /**
     * @return Collection|Participant[]
     */
    public function getParticipations (): Collection {
        return $this->participations;
    }
    public function addParticipation (Participant $participation): self {
        if (!$this->participations->contains($participation)) {
            $this->participations[] = $participation;
            $participation->setUser($this);
        }

        return $this;
    }
    public function removeParticipation (Participant $participation): self {
        if ($this->participations->contains($participation)) {
            $this->participations->removeElement($participation);
            // set the owning side to null (unless already changed)
            if ($participation->getUser() === $this) {
                $participation->setUser(null);
            }
        }

        return $this;
    }

    /**
     * @return Collection|Comment[]
     */
    public function getComments (): Collection {
        return $this->comments;
    }
    public function addComment (Comment $comment): self {
        if (!$this->comments->contains($comment)) {
            $this->comments[] = $comment;
            $comment->setUser($this);
        }

        return $this;
    }
    public function removeComment (Comment $comment): self {
        if ($this->comments->contains($comment)) {
            $this->comments->removeElement($comment);
            // set the owning side to null (unless already changed)
            if ($comment->getUser() === $this) {
                $comment->setUser(null);
            }
        }

        return $this;
    }
}
